# DNA

Dart Native Access lets you deal with native libraries from Dart with zero lines of C/C++ code.
Just pure Dart.

## How to ...

For instance, you want to call `getppid` function from `libc` library on Linux.

* Find method signature in documentation:

```C
pid_t getpid(void);
```

* Map parameters types and return type to _C_ data types:

```C
typedef int	pid_t;
```

* Map _C_ data types to Dart types and DNA type constants. See **Type mapping**.


* Define class `Libc` with annotation `@Library`:
 
```Dart
@Library('libc.so.6')
class Libc {
  
}
```
`libc.so.6` is library name 

* Define method `getppid`. Annotate parameters with `@Param` and method with `@Method` with corresponding type constants:

```Dart
class Libc {
  int getpid() { ... }
}
```
In this example method has no parameters.

* Finish with boilerplate code:

```Dart
class Libc {
  dynamic lib = new DynamicLibrary<Libc>();  
  
  int getpid() { return lib.getpid(); }
}

```

* Use it

```Dart
Libc libc = new Libc();
var processId = libc.getpid();
```
For full code see **Examples**.

## How to define library

```Dart
@Library('name')
class Library {
    dynamic lib = new DynamicLibrary<Library>(); //It's required yet
}
```
 _name_ is the library name. For instance: _libname.so_ or _name.dll_ or _libname.dylib_.
 
Please see documentation for the target OS to know where dynamic linker searches for the library.
* Linux - [dlopen](http://man7.org/linux/man-pages/man3/dlopen.3.html)
* Mac OS - [dlopen](https://developer.apple.com/legacy/library/documentation/Darwin/Reference/ManPages/man3/dlopen.3.html)
* Windows - [LoadLibrary](https://msdn.microsoft.com/en-us/library/windows/desktop/ms684175.aspx)

> FIXME: Symlinks aren't supported

## How to define method

```Dart
@Method(C_XYZ) T method(
  @Param(C_XYZ) T1 inParam, 
  @Param(C_XYZ, out: true) Ref<T2> outParam) { 
    return lib.method(inParam, outParam); 
}
```
_C_XYZ_ is _C_ type constant. See **Type mapping**.

## How to define structure

```Dart
@Struct()
class Struct {
    @Field(C_XYZ) T field;
}
```

## Type mapping

| _C_    | Dart   |Constant|In/Out  |Comments|
|-----   |-----   |-----   |-----   |-----|
|void    |`void`  |VOID    |In      ||
|char    |`int`   |C_CHAR  |In      ||
|short   |`int`   |C_SHORT |In      ||
|int     |`int`   |C_INT   |In      ||
|long    |`int`   |C_LONG  |In      ||
|longlong|`int`   |C_LONGLONG|In    ||
|bool    |`bool`  |C_BOOL  |In      ||
|float   |`double`|C_FLOAT |In      ||
|double  |`double`|C_DOUBLE|In      ||
|---     |---     |---     |---     |---|
|{type} *|`int`   |C_POINTER|In     |Raw pointer value.|
|{type} *|`TypedData`|TYPEDDATA|In\Out|TypeData must be initialized and have expected by callee size.|
|---     |---     |---     |---     |---|
|struct *|`T`     |C_STRUCT|In      | T is struct class. See **Define structure**.|
|struct *|`Ref<T>`|C_STRUCT|Out     |T is struct class. See **Define structure**.|
|---     |---     |---     |---     ||
|char *  |`String`|C_STRING|In      ||
|char *  |`Ref<String>`|C_STRING|Out||
|char ** |`List<String>`|LISTSTRING|In||
|{integer type} *|`List<int>`|C_{type}|In||
|{integer type} *|`Ref<List<int>>`|C_{type}|Out||
|{decimal type} *|`List<double>`|C_{type}|In||
|{decimal type} *|`Ref<List<double>>`|C_{type}|Out||

### Out parameters

Out parameters must be marked as `@Param( , out: true)`

Dart doesn't support parameters by reference. 
It's good but it's the cause of one challenge: callee cannot override parameter object.
For instance, `String` parameter cannot be modified inside called method.
 
Therefore, it's needed to use wrapper object to emulate out parameters: `Ref<T>`.

> FIXME: Describe pointers, in/out parameters, String parameters and TypedData parameters

## FAQ

> I see `Failed assertion: 'libraryPointer != 0`

Dynamic linker cannot find the library. Check the library name.

> I see `Failed assertion: 'methodPointer != 0`

Dynamic linker cannot find method in the library. Check the method name. Check that library exports the method.  

> My library is cross platform and has different name on each platform.

See example.
```Dart
@Library('libname.so', 'name.dll')
class Library {
    dynamic lib = new DynamicLibrary<Library>();
}
```

> My library doesn't export functions or/and uses custom logic to get function pointer

See example.
```Dart
@Library('name.dll')
class Library {
  dynamic lib = new DynamicLibrary<Library>(getMethodPointer);

  static int getMethodPointer(DynamicLibrary that, Invocation invocation) {
    return ... //raw function pointer here
  }
}
```

> Why is it needed to define such strange method body `{ return lib.name(); }` ?

Because it is impossible to create class at runtime in Dart.


## Examples

### Linux

#### libc library
```Dart
    @Library('libc.so.6')
    class Libc {
      dynamic lib = new DynamicLibrary<Libc>();
    
      int getpid() { return lib.getpid(); }      
      int getuid() { return lib.getuid(); }       
               
      void srand(@Param(C_UINT) int seed) { return lib.srand(seed); }      
      int rand() { return lib.rand(); }
    }
    
    void main(){
      Libc libc = new Libc();
      var processId = libc.getpid();
      print('pid $processId');
    
      var userId = libc.getuid();
      print('user id ${userId}');
    
      libc.srand(0xDEADBEEF);
      var rand = libc.rand();
      print('random $rand');    
    }
```
### Windows   
 
#### kenel32 library
```Dart
    @Library('Kernel32.dll')
    class Kernel32 {
      dynamic lib = new DynamicLibrary<Kernel32>();
    
      int GetCurrentProcess() { return lib.GetCurrentProcess(); }
      int GetProcessId(@Param(C_INT) int process) { return lib.GetProcessId(process); }
      
      int GetModuleFileNameA(@Param(C_INT) int process, @Param(C_STRING, out: true) Ref<String> imageFileName, 
        @Param(C_INT) int size) { return lib.GetModuleFileNameA(process, imageFileName, size); }
        
      bool GetProcessWorkingSetSize(@Param(C_INT) int process, 
        @Param(C_POINTER) Ref<int> minimumWorkingSetSize,
        @Param(C_POINTER) Ref<int> maximumWorkingSetSize) { return lib.GetProcessWorkingSetSize(process, minimumWorkingSetSize, maximumWorkingSetSize); }
    
      int GetLastError() { return lib.GetLastError(); }
    }
    
    void main(){
      Kernel32 kernel32 = new Kernel32();
      var process = kernel32.GetCurrentProcess();
      var processId = kernel32.GetProcessId(process);
      print('pid $processId = ${io.pid}');
    
      var min = new Ref<int>(0);
      var max = new Ref<int>(0);
      var result = kernel32.GetProcessWorkingSetSize(process, min, max);
      print('working set min ${min.value} max ${max.value}');
    
      var name = new Ref(new String.fromCharCodes(new List.filled(64, 0)));
      result = kernel32.GetModuleFileNameA(0, name, name.value.length);
      print('module name  ${name.value} result $result error ${kernel32.GetLastError()}');
    
    }
```
Please find more examples in `examples/`.
    
## Requirements

* Linux 64-bit
* Windows 32-bit and 64-bit

## To Do
* Improve performance
* A lot of things aren't supported yet 
  * Platform independent types: int8, int16, int32, int64 etc.
  * Function pointers to Dart methods
  * Custom structure align
  * Unions
  * ...
* Get rid off template code in library class
* Mac OS support  
* Refactor and clean up code
