//  Copyright 2016 root.ext@gmail.com
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import 'dart:mirrors';
import 'dart:typed_data';

import 'native.dart';
import 'util.dart';
import 'meta.dart';
import 'model/field.dart';

TypedData encodeStruct(dynamic object) {
  var totalSize = structSize(object);

  var offset = 0;
  TypedData data = new Uint8List(totalSize);
  ByteData view = new ByteData.view(data.buffer);
  iterateFields(object, (field, type, fieldMirror) {
    var size = fieldSize(type);
    offset += align(offset, size);

    if (isInteger(type)) {
      setInt(view, size, offset, field != null ? (field as int) : 0);
    } else if (isDecimal(type)) {
      setFloat(view, size, offset, field != null ? (field as double) : 0.0);
    }

    switch (type) {
      case C_BOOL:
        setInt(view, size, offset, (field as bool) ? 1 : 0);
        break;
      case C_STRING:
        if (field == null) {
          setInt(view, size, offset, 0);
          break;
        }
        TypedData data = new Uint8List.fromList((field as String).codeUnits);
        setInt(view, size, offset, getPointer(data));
        break;
      case C_STRUCT:
        TypedData data = encodeStruct(field);
        setInt(view, size, offset, getPointer(data));
        break;
    }

    offset += size;
  });

  //print('encode $data ${data.toList().length}');
  return data;
}

dynamic decodeStruct(TypedData data, dynamic object) {
  //print('decode $data');

  ByteData view = new ByteData.view(data.buffer);
  InstanceMirror classMirror = reflect(object);

  var offset = 0;
  iterateFields(object, (field, type, fieldMirror) {
    var size = fieldSize(type);
    offset += align(offset, size);

    dynamic value;
    if (isInteger(type)) {
      value = getInt(view, size, offset);
    } else if (isDecimal(type)) {
      value = getFloat(view, size, offset);
    }
    switch (type) {
      case C_BOOL:
        value = getInt(view, size, offset) == 1;
        break;
      case C_STRING:
        value = "";
        break;
    }

    classMirror.setField(fieldMirror.simpleName, value);
    offset += size;
  });
  return object;
}

TypedData encodeList(List object, int type) {
  var size = fieldSize(type);

  TypedData data;
  if (isInteger(type)) {
    switch (size) {
      case 1:
        data = new Uint8List.fromList(object);
        break;
      case 2:
        data = new Uint16List.fromList(object);
        break;
      case 4:
        data = new Uint32List.fromList(object);
        break;
      case 8:
        data = new Uint64List.fromList(object);
        break;
    }
  } else if (isDecimal(type)) {
    switch (size) {
      case 4:
        data = new Float32List.fromList(object);
        break;
      case 8:
        data = new Float64List.fromList(object);
        break;
    }
  }

  //print('encode $data ${data.toList().length}');
  return data;
}

List decodeList(TypedData data, int type) {
  //print('decode $data');

  var size = fieldSize(type);
  List list;

  if (isInteger(type)) {
    switch (size) {
      case 1:
        list = (data as Uint8List).toList();
        break;
      case 2:
        list = (data as Uint16List).toList();
        break;
      case 4:
        list = (data as Uint32List).toList();
        break;
      case 8:
        list = (data as Uint64List).toList();
        break;
    }
  } else if (isDecimal(type)) {
    switch (size) {
      case 4:
        list = (data as Float32List).toList();
        break;
      case 8:
        list = (data as Float64List).toList();
        break;
    }
  }

  return list;
}

int fieldSize(int type) {
  int size;
  switch (type) {
    case C_STRING:
    case C_STRUCT:
      size = ctypeToSize(C_POINTER);
      break;
    default:
      size = ctypeToSize(type);
      break;
  }
  return size;
}

int structSize(dynamic object) {
  var totalSize = 0;
  var maxFieldSize = 0;
  iterateFields(object, (field, type, fieldMirror) {
    var size = fieldSize(type);
    totalSize += align(totalSize, size);
    totalSize += size;

    maxFieldSize = size > maxFieldSize ? size : maxFieldSize;
  });
  totalSize += align(totalSize, maxFieldSize);

  return totalSize;
}

void iterateFields(dynamic struct,
    dynamic process(dynamic field, int type, DeclarationMirror fieldMirror)) {
  ClassMirror classMirror = reflect(struct).type;

  var fields = classMirror.declarations.values.where((d) => true);

  for (var fieldMirror in fields) {
    var fieldTag = findDeclTag(fieldMirror, Field);
    if (fieldTag == null) continue;

    var fieldType = fieldTag != null
        ? (fieldTag.getField(#type).reflectee as int)
        : dartTypeToCtype(reflectType(fieldMirror.runtimeType));
    var field = reflect(struct).getField(fieldMirror.simpleName).reflectee;

    process(field, fieldType, fieldMirror);
  }
}

int align(int offset, int size) {
  return offset % size != 0 ? (size - offset % size) : 0;
}

List<int> sizes = getSizes();

int ctypeToSize(int type) {
  switch (type) {
    case C_CHAR:
    case C_UCHAR:
      return sizes[0];
    case C_SHORT:
    case C_USHORT:
      return sizes[1];
    case C_INT:
    case C_UINT:
      return sizes[2];
    case C_LONG:
    case C_ULONG:
      return sizes[3];
    case C_LONGLONG:
    case C_ULONGLONG:
      return sizes[4];
    case C_BOOL:
      return sizes[5];
    case C_FLOAT:
      return sizes[6];
    case C_DOUBLE:
      return sizes[7];
    case C_POINTER:
      return sizes[8];
    case C_STRING:
      return sizes[8];
    case C_INT8:
      return 1;
    case C_INT16:
      return 2;
    case C_INT32:
      return 4;
    case C_INT64:
      return 8;
    default:
      return 0;
  }
}

bool isInteger(int type) {
  return (type >= C_CHAR) && (type <= C_INT64) && (type != C_BOOL) ||
      (type == C_POINTER);
}

bool isDecimal(int type) {
  return (type >= C_FLOAT) && (type <= C_DOUBLE);
}
